﻿//TCP Client/Server 
//original code based on http://www.techcoil.com/blog/implementing-client-server-communication-using-serialization-and-tcpip-in-c/
//
//modified by Simon Carter (s1cart3r@gmail.com)
//
//Redistribution and use in source and binary forms are permitted
//provided that the above copyright notice and this paragraph are
//duplicated in all such forms and that any documentation,
//advertising materials, and other materials related to such
//distribution and use acknowledge that the software was developed
//by techcoil.com and Simon Carter.  The name of the
//techcoil.com and Simon Carter may not be used to endorse or promote products derived
//from this software without specific prior written permission.
//THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
//IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
//WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//change history
// version 1.0 - initial release 19/05/2013
//


using System;
using System.IO;
using System.Runtime.Serialization;

namespace TCPMessageServer
{
    // Enumeration that define the type of message
    public enum MessageType
    {
        /// <summary>
        /// Error Message
        /// </summary>
        Error,

        /// <summary>
        /// Warning Message
        /// </summary>
        Warning,

        /// <summary>
        /// Information Message
        /// </summary>
        Info,

        /// <summary>
        /// Acknowledgement
        /// </summary>
        Acknowledge,

        /// <summary>
        /// Command message
        /// </summary>
        Command,

        /// <summary>
        /// Broadcasts message to all clients
        /// </summary>
        Broadcast,

        /// <summary>
        /// File Transfer message
        /// </summary>
        File,

        /// <summary>
        /// Sends a message to an individual user
        /// </summary>
        User
    }

    [Serializable]
    public class Message
    {
        #region Private Members

        private MessageType _messageType;
        private string _messageTitle;
        private string _messageContents;
        private string _clientID;

        #endregion Private Members

        #region Constructors

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="title">Title of message</param>
        /// <param name="contents">Content of message</param>
        /// <param name="type">message type</param>
        public Message(string title, string contents, MessageType type)
        {
            this.Title = title;
            this.Contents = contents;
            this.Type = type;
        }

        #endregion Constructors

        #region Properties

        /// <summary>
        /// Get Client ID of the connection as set by the server
        /// </summary>
        public string ClientID
        {
            get
            {
                return (_clientID);
            }
        }

        /// <summary>
        /// Get/Set the type of message
        /// </summary>
        public MessageType Type
        {
            get
            {
                return this._messageType;
            }

            set
            {
                this._messageType = value;
            }
        }

        /// <summary>
        /// Get/Set the title of the message
        /// </summary>
        public string Title
        {
            get
            {
                return this._messageTitle;
            }

            set
            {
                this._messageTitle = value;
            }
        }

        /// <summary>
        /// Get/Set the content of the message
        /// </summary>
        public string Contents
        {
            get
            {
                return this._messageContents;
            }

            set
            {
                this._messageContents = value;
            }
        }

        #endregion Properties

        #region Internal Methods

        /// <summary>
        /// Methods used internally to set the client id
        /// </summary>
        /// <param name="ClientID">New Client ID</param>
        internal void SetClientID(string ClientID)
        {
            _clientID = ClientID;
        }

        #endregion Internal Methods

        #region Static Methods

        /// <summary>
        /// Creates a Command Message
        /// </summary>
        /// <param name="command">Command to be passed</param>
        /// <param name="parameters">Optional parameters to be passed in the body of the command Message</param>
        /// <returns>Message Object</returns>
        public static Message Command(string command, string parameters = "")
        {
            return (new Message(command, parameters, MessageType.Command));
        }


        /// <summary>
        /// Creates a Error Message for Feature Not Supported
        /// </summary>
        /// <param name="contents">Optional contents for message</param>
        /// <returns></returns>
        public static Message FeatureNotSupported(string contents = "")
        {
            return (new Message("FEATURE_NOT_SUPPORTED", contents, MessageType.Error));
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public static Message FileToStringMessage(string fileName)
        {
            Message Result = new Message(Path.GetFileName(fileName), "", MessageType.File);

            using (FileStream fs = new FileStream(fileName, FileMode.Open))
            {
                using (BinaryReader br = new BinaryReader(fs))
                {
                    byte[] bin = br.ReadBytes(Convert.ToInt32(fs.Length));
                    Result.Contents = Convert.ToBase64String(bin);
                }
            }

            return (Result);
        }

        #endregion Static Methods
    }


}
