﻿//TCP Client/Server 
//original code based on http://www.techcoil.com/blog/implementing-client-server-communication-using-serialization-and-tcpip-in-c/
//
//modified by Simon Carter (s1cart3r@gmail.com)
//
//Redistribution and use in source and binary forms are permitted
//provided that the above copyright notice and this paragraph are
//duplicated in all such forms and that any documentation,
//advertising materials, and other materials related to such
//distribution and use acknowledge that the software was developed
//by techcoil.com and Simon Carter.  The name of the
//techcoil.com and Simon Carter may not be used to endorse or promote products derived
//from this software without specific prior written permission.
//THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
//IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
//WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//change history
// version 1.0 - initial release 19/05/2013
//

using System;
using System.Net;
using System.Net.Sockets;
using System.Runtime.Serialization.Formatters.Binary;
using System.Threading;
using System.IO;

using TCPMessageServer.Utilities;

namespace TCPMessageServer
{
    /// <summary>
    /// Client object used for receiving messages
    /// </summary>
    public class MessageClient
    {
        #region Private / Internal Membes

        private int _port;
        private string _server;
        private Thread _listenThread;
        private bool _running;
        private string _clientID;
        private bool _ignoreBroadcasts;
        private bool _timeout;
        private FileBufferSize _bufferSize = FileBufferSize.Size1024;
        private bool _loggedIn = false;

        internal TcpClient _tcpClient;

        #endregion Private / Internal Members

        #region Constructors

        /// <summary>
        /// Constructor, used for localhost only
        /// </summary>
        /// <param name="port">Port to connect to</param>
        public MessageClient(int port)
        {
            _port = port;
            _server = "localhost";
            _running = false;
            _ignoreBroadcasts = false;
        }

        /// <summary>
        /// Constructor, specify server ip address/name and port
        /// </summary>
        /// <param name="server">Server Name or IP Address</param>
        /// <param name="port">Port to connect to</param>
        public MessageClient(string server, int port)
        {
            this._port = port;
            _server = server;
            _running = false;
        }

        #endregion Constructors

        #region Public Methods

        /// <summary>
        /// Initialises the client and starts listening for messages from the server
        /// </summary>
        public void StartListening()
        {
            try
            {
                RaiseConnected();
                _timeout = false;

                lock (this)
                {
                    _tcpClient = new TcpClient(_server, _port);
                    _listenThread = new Thread(new ThreadStart(ListenForMessage));
                    _listenThread.Start();
                }
            }
            catch (Exception err)
            {
                if (!HandleClientException(err))
                    throw;
            }
        }

        /// <summary>
        /// Disconnects and stops listening for messages
        /// </summary>
        public void StopListening()
        {
            lock (this)
            {
                if (_running)
                {
                    if (IsConnected)
                        sendMessage(new Message("CLOSING", "", MessageType.Command));

                    RaiseDisconnected();
                    _tcpClient.Close();
                }

                _tcpClient = null;
            }
        }

        /// <summary>
        /// Send a message to the server
        /// </summary>
        /// <param name="message">Message to be sent</param>
        public void SendMessage(Message message)
        {
            lock (this)
            {
                if (_running)
                {
                    sendMessage(message);
                }
            }
        }

        /// <summary>
        /// Sends a file to the server
        /// </summary>
        /// <param name="FileName">File to be sent</param>
        public void SendFile(string FileName)
        {
            FileInfo f = new FileInfo(FileName);

            if (f.Length > 20971520) //20 mb limit
            {
                RaiseError(new Exception("File exceeds maximum file size"));
            }
            else
            {
                sendMessage(Message.FileToStringMessage(FileName));
                //sendMessage(new Message("SEND_FILE", String.Format("{0}${1}", FileName, f.Length), MessageType.File)); // old way
                RaiseFileReceived(this, new TransferFileEventArgs(FileName));
            }
        }

        /// <summary>
        /// Instructs the server to send a file
        /// 
        /// Server must manage its own path system as only the file name will be sent
        /// </summary>
        /// <param name="FileName">Name of file to be sent by server</param>
        public void ReceiveFile(string FileName)
        {
            sendMessage(new Message("REQUEST_FILE", FileName, MessageType.File));
        }

        #endregion Public Methods

        #region Private Methods

        /// <summary>
        /// Handles exceptions when handling client connections
        /// </summary>
        /// <param name="error">Exception raised</param>
        /// <param name="clientID">clientID when error raised</param>
        /// <returns></returns>
        private bool HandleClientException(Exception error)
        {
            // connection is about to close
            if (_timeout)
            {
                RaiseDisconnected();
                return (true);
            }

            if (error.Message.Contains("does not contain a valid BinaryHeader"))
            {
                //we have sent/received junk, disconnect the client
                RaiseDisconnected();
                _tcpClient.Close();
                return (true);
            }

            if (error.Message.Contains("An established connection was aborted by the software in your host machine"))
            {
                RaiseDisconnected();
                return (true);
            }

            if (error.Message.Contains("An existing connection was forcibly closed by the remote host"))
            {
                RaiseDisconnected();
                return (true);
            }

            if (error.Message.Contains("End of Stream encountered before parsing was completed"))
            {
                RaiseDisconnected();
                return (true);
            }

            if (error.Message.Contains("The operation is not allowed on non-connected sockets."))
            {
                RaiseDisconnected();
                return (true);
            }

            //trying to send message when connection closed
            if (error.Message.Contains("Unable to read data from the transport connection: A blocking operation was interrupted by a call to WSACancelBlockingCall"))
            {
                if (_running)
                {
                    RaiseDisconnected();
                }

                return (true);
            }

            return (RaiseError(error));
        }

        /// <summary>
        /// Listener used by thread to listen for messages
        /// </summary>
        private void ListenForMessage()
        {
            try
            {
                while (_running)
                {
                    // Block until an instance Message is received
                    BinaryFormatter bFormatter = new BinaryFormatter();
                    Message message = (Message)bFormatter.Deserialize(this._tcpClient.GetStream());

                    //have we stopped running?
                    if (!_running)
                        return;

                    if (MessageReceived != null && message != null)
                    {
                        switch (message.Type)
                        {
                            case MessageType.Command:
                                //some command messages are handled internally
                                switch (message.Title)
                                {
                                    case "NEWCLIENTID":
                                        _clientID = message.Contents;
                                        RaiseClientIDChanged();
                                        break;
                                    case "SERVERSTOPPING":
                                        StopListening();
                                        break;
                                    case "LOGIN_REQUIRED":
                                        string loginDetails = RaiseLoginRequired(message.Contents);

                                        sendMessage(new Message("LOGIN", loginDetails, MessageType.Command));
                                        break;
                                    case "LOGIN_FAILED":
                                        StopListening();
                                        RaiseLoginFailed();
                                        break;
                                    case "LOGGED_IN":
                                        RaiseLoggedIn();
                                        break;
                                    case "TIMEOUT":
                                        _timeout = true;
                                        break;
                                    case "CONNECTION_REFUSED":
                                        RaiseConnectionRefused();
                                        break;
                                    default: //not an internal message process as normal
                                        message.SetClientID(_clientID);
                                        MessageReceived(this, message);
                                        break;
                                }

                                break;
                            case MessageType.File:
                                string newFileName = RaiseFileReceive(message.Title);
                                //FileTransfer transfer = new FileTransfer();
                                //transfer.FileReceived += RaiseFileReceived;
                                //transfer.ProcessClientFiles(this, message);

                                byte[] rebin = Convert.FromBase64String(message.Contents);
                                using (FileStream fs = new FileStream(newFileName, FileMode.Create))
                                {
                                    using (BinaryWriter bw = new BinaryWriter(fs))
                                        bw.Write(rebin);
                                }

                                RaiseFileReceived(this, new TransferFileEventArgs(newFileName));
                                MessageReceived(this, message);

                                continue;
                            default:

                                // any other message forward onto event handler
                                message.SetClientID(_clientID);
                                MessageReceived(this, message);
                                break;
                        }
                    }
                }
            }
            catch (Exception err)
            {
                if (!HandleClientException(err))
                    throw;
            }
        }

        #endregion Private Methods

        #region Internal Methods

        /// <summary>
        /// Sends the message to the server
        /// </summary>
        /// <param name="message">Message to be sent</param>
        internal void sendMessage(Message message)
        {
            try
            {
                if (_tcpClient.Connected)
                {
                    message.SetClientID(_clientID);
                    BinaryFormatter bFormatter = new BinaryFormatter();
                    bFormatter.Serialize(_tcpClient.GetStream(), message);
                }
            }
            catch (Exception err)
            {
                if (!HandleClientException(err))
                    throw;
            }
        }

        #endregion Internal Methods

        #region Properties

        /// <summary>
        /// Determines whether broadcast messages are ignored or not
        /// </summary>
        public bool IgnoreBroadcasts
        {
            get
            {
                return (_ignoreBroadcasts);
            }

            set
            {
                _ignoreBroadcasts = value;
                SendMessage(new Message("IGNORE_BROADCAST", _ignoreBroadcasts.ToString(), MessageType.Command));
            }
        }

        /// <summary>
        /// Returns true if the client is running and connected to the server
        /// </summary>
        public bool IsRunning
        {
            get
            {
                return (_running);
            }
        }

        /// <summary>
        /// Returns the unique Client ID as identified by the server
        /// </summary>
        public string ClientID
        {
            get
            {
                return (_clientID);
            }
        }

        /// <summary>
        /// Returns the server the client is connected to or ("Not Connected")
        /// </summary>
        public string Server
        {
            get
            {
                if (IsRunning)
                    return (_server);
                else
                    return ("Not Connected");
            }
        }

        /// <summary>
        /// Determines whether the client is connected to the server or not
        /// </summary>
        public bool IsConnected
        {
            get
            {
                if (_tcpClient == null)
                    return (false);
                else
                    return (_tcpClient.Connected);
            }
        }

        /// <summary>
        /// Secifies the buffer size for sending/receiving files files
        /// </summary>
        public FileBufferSize BufferSize
        {
            get
            {
                return (_bufferSize);
            }

            set
            {
                _bufferSize = value;

                //update server with new buffer size
                sendMessage(new Message("SET_BUFFER_SIZE", Convert.ToString((UInt32)value), MessageType.Command));
            }
        }

        /// <summary>
        /// Indicates whether the client is logged into the server or not
        /// </summary>
        public bool LoggedIn
        {
            get
            {
                return (_loggedIn);
            }
        }

        #endregion Properties

        #region events

        #region Private Methods

        /// <summary>
        /// Raises client connected event
        /// </summary>
        private void RaiseConnected()
        {
            _running = true;

            if (Connected != null)
                Connected(this, EventArgs.Empty);
        }

        /// <summary>
        /// Raises client disconnected event
        /// </summary>
        private void RaiseDisconnected()
        {
            _running = false;

            if (Disconnected != null)
                Disconnected(this, EventArgs.Empty);
        }

        /// <summary>
        /// Raises client ID changed event
        /// </summary>
        private void RaiseClientIDChanged()
        {
            if (ClientIDChanged != null)
                ClientIDChanged(this, EventArgs.Empty);
        }

        /// <summary>
        /// Raises Internal error event handler
        /// </summary>
        /// <param name="error">Error that occured</param>
        /// <returns>bool true to continue false to stop the server</returns>
        private bool RaiseError(Exception error)
        {
            ErrorEventArgs args = new ErrorEventArgs(error);

            if (OnError != null)
                OnError(this, args);

            return (args.Continue);
        }

        /// <summary>
        /// Raise LoginRequired Event
        /// </summary>
        /// <param name="passwordCode">Random encryption password created by the server</param>
        /// <returns></returns>
        private string RaiseLoginRequired(string passwordCode)
        {
            string Result = "$";

            ClientLoginArgs args = new ClientLoginArgs();

            if (ClientLogin != null)
            {
                ClientLogin(this, args);

                if (args.Username != null || args.Password != null)
                    Result = String.Format("{0}${1}", args.Username, StringCipher.Encrypt(args.Password, passwordCode));
            }

            return (Result);
        }

        /// <summary>
        /// Raises login failed event
        /// </summary>
        private void RaiseLoginFailed()
        {
            if (ClientLoginFailed != null)
                ClientLoginFailed(this, EventArgs.Empty);
        }

        /// <summary>
        /// Raises login success event
        /// </summary>
        private void RaiseLoggedIn()
        {
            //show that we are logged in
            _loggedIn = true;

            if (ClientLoginSuccess != null)
                ClientLoginSuccess(this, EventArgs.Empty);
        }

        /// <summary>
        /// Raises file received event
        /// 
        /// Used to monitor progres of file being sent/received
        /// </summary>
        /// <param name="sender">sender object</param>
        /// <param name="e">ReceiveFileEventArgs event arguments</param>
        private void RaiseFileReceived(object sender, TransferFileEventArgs e)
        {
            if (FileReceived != null)
                FileReceived(this, e);
        }

        /// <summary>
        /// Raises connection refused event
        /// </summary>
        private void RaiseConnectionRefused()
        {
            if (ConnectionRefused != null)
                ConnectionRefused(this, EventArgs.Empty);
        }

        #endregion Private Methods

        #region Internal Methods

        /// <summary>
        /// Raises a FileReceive event
        /// </summary>
        /// <param name="FileName">Filename being received</param>
        /// <returns>Client should return a fully qualified path/file name of the file to be returned</returns>
        internal string RaiseFileReceive(string FileName)
        {
            TransferFileEventArgs args = new TransferFileEventArgs(FileName);

            if (FileReceive != null)
                FileReceive(this, args);

            return (args.FileName);
        }

        #endregion Internal Methods

        public event MessageReceivedEventHandler MessageReceived;
        public event EventHandler ClientIDChanged;
        public event EventHandler Connected;
        public event EventHandler Disconnected;
        public event ErrorEventHandler OnError;
        public event ClientLoginHandler ClientLogin;
        public event EventHandler ClientLoginFailed;
        public event EventHandler ClientLoginSuccess;
        public event FileReceivedHandler FileReceive;
        public event FileReceivedHandler FileReceived;
        public event EventHandler ConnectionRefused;

        #endregion events
    }
}
