﻿//TCP Client/Server 
//original code based on http://www.techcoil.com/blog/implementing-client-server-communication-using-serialization-and-tcpip-in-c/
//
//modified by Simon Carter (s1cart3r@gmail.com)
//
//Redistribution and use in source and binary forms are permitted
//provided that the above copyright notice and this paragraph are
//duplicated in all such forms and that any documentation,
//advertising materials, and other materials related to such
//distribution and use acknowledge that the software was developed
//by techcoil.com and Simon Carter.  The name of the
//techcoil.com and Simon Carter may not be used to endorse or promote products derived
//from this software without specific prior written permission.
//THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
//IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
//WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//change history
// version 1.0 - initial release 19/05/2013
//

using System;
using System.Collections.Generic;
using System.Net;
using System.IO;
using System.Net.Sockets;
using System.Runtime.Serialization.Formatters.Binary;
using System.Threading;

using TCPMessageServer.Utilities;

namespace TCPMessageServer
{
    /// <summary>
    /// TCP Server object
    /// </summary>
    public class MessageServer
    {
        #region Private Members

        private int _port;
        private TcpListener _tcpListener;
        private Dictionary<string, ConnectedClient> _clientsDictionary;
        private Thread _connectionThread;
        private bool _running;
        private string activeNewClientID = "";
        private Int64 _nextClientID = 0;
        private bool _loginRequired;
        private ushort _maxClientConnections;
        private ushort _clientTimeout = 0;

        #endregion Private Members

        #region Constructors

        /// <summary>
        /// Create a message server that listens on the indicated port
        /// </summary>
        /// <param name="port">Port to connect to</param>
        public MessageServer(int port)
        {
            _port = port;
            _tcpListener = new TcpListener(IPAddress.Any, port);
            _clientsDictionary = new Dictionary<string, ConnectedClient>();
            _running = false;
            _loginRequired = false;
            _maxClientConnections = 0;
            MaximumFileSize = 5242880; //5MB default limit
        }

        #endregion Constructors

        #region Public Methods

        /// <summary>
        /// Sends a message to all connected clients (broadcast)
        /// </summary>
        /// <param name="message">Message to be sent</param>
        public void SendMessage(Message message, bool allClients = true)
        {
            string activeClientID = "";
            ConnectedClient activeClient = null;
            try
            {
                //copy the clients dictionary in case its modified whilst iterating
                Dictionary<string, ConnectedClient> clients = new Dictionary<string, ConnectedClient>(_clientsDictionary);

                foreach (KeyValuePair<string, ConnectedClient> entry in clients)
                {
                    activeClientID = entry.Key;
                    activeClient = entry.Value;

                    // If client had connected to the message server
                    if (allClients || (!allClients && clients.ContainsKey(activeClientID) && activeClientID == message.ClientID))
                    {
                        try
                        {
                            // send message to client
                            if (activeClient != null && activeClient.Client.Connected)
                            {
                                // does the client need to login?
                                if (!_loginRequired || (_loginRequired && clients[activeClientID].LoggedIn))
                                {
                                    // is the client ignoring broadcast messages
                                    if (message.Type != MessageType.Broadcast || (!clients[activeClientID].IgnoreBroadcastMessages && message.Type == MessageType.Broadcast))
                                    {
                                        sendMessage(activeClientID, clients[activeClientID].Client, message);
                                    }
                                }
                            }
                        }
                        catch (Exception)
                        {
                            RaiseClientDisconnected(ClientAddress(activeClient.Client), activeClientID);
                            message.Type = MessageType.Command;
                            message.Title = "CLOSING";
                            message.SetClientID(activeClientID);
                            ProcessCommand(message);
                        }
                    }
                }

                clients = null;
                activeClient = null;
            }
            catch (Exception err)
            {
                if (!HandleClientException(err, activeClientID, activeClient.Client))
                    throw;
            }
        }

        /// <summary>
        /// Starts the server
        /// </summary>
        public void Start()
        {
            if (!_running)
            {
                _tcpListener.Start();
                RaiseStarted();

                _connectionThread = new Thread
                    (new ThreadStart(ListenForClientConnections));
                _connectionThread.Start();
                _connectionThread.Name = "MessageServer Connection Thread";

                Thread maintenance = new Thread(new ThreadStart(MaintenanceThread));
                maintenance.Name = "MessageServer Maintenance Thread";
                maintenance.Priority = ThreadPriority.Lowest;
                maintenance.Start();
            }
        }

        /// <summary>
        /// Stops the Server
        /// </summary>
        public void Stop()
        {
            if (_running)
            {
                SendMessage(Message.Command("SERVERSTOPPING"));
                RaiseStopped();
                _tcpListener.Stop();

                lock (_clientsDictionary)
                {
                    _clientsDictionary.Clear(); //clear all connections
                }
            }
        }

        /// <summary>
        /// Returns the logged in user name for a Client
        /// </summary>
        /// <param name="ClientID">Client ID of user logged in</param>
        /// <returns>Username of logged in client</returns>
        public string UserName(string ClientID)
        {
            lock (_clientsDictionary)
            {
                return (_clientsDictionary[ClientID].LoginName);
            }
        }

        /// <summary>
        /// retrieves user defined data for the client connection
        /// </summary>
        /// <param name="ClientID">Client ID</param>
        /// <returns>User defined object to be returned</returns>
        public object UserData(string ClientID)
        {
            lock (_clientsDictionary)
            {
                return (_clientsDictionary[ClientID].UserData);
            }
        }

        /// <summary>
        /// Sets user defined data for the client connection
        /// </summary>
        /// <param name="ClientID">Client ID</param>
        /// <param name="Data">Data to be saved with client connection</param>
        public void UserData(string ClientID, object Data)
        {
            lock (_clientsDictionary)
            {
                _clientsDictionary[ClientID].UserData = Data;
            }
        }

        #endregion Public Methods

        #region Properties

        /// <summary>
        /// Indicates wether the server is running or not
        /// </summary>
        public bool Running
        {
            get
            {
                return (_running);
            }
        }

        /// <summary>
        /// Indicates wether clients need to login
        /// </summary>
        public bool LoginRequird
        {
            get
            {
                return (_loginRequired);
            }

            set
            {
                _loginRequired = value;
            }
        }

        /// <summary>
        /// Maximum number of Client Connections
        /// </summary>
        public ushort MaxClientConnections
        {
            get
            {
                return (_maxClientConnections);
            }

            set
            {
                _maxClientConnections = value;
            }
        }

        /// <summary>
        /// Determintes the number of seconds the client received/sent data before timeout (0 = no limit)
        /// </summary>
        public ushort ClientTimeOut
        {
            get
            {
                return (_clientTimeout);
            }

            set
            {
                _clientTimeout = value;
            }
        }

        /// <summary>
        /// Determines wether the server accepts files or not
        /// </summary>
        public bool AcceptFiles
        {
            get;
            set;
        }

        /// <summary>
        /// Maximum file size in bytes that the server will accept
        /// </summary>
        public ulong MaximumFileSize
        {
            get;
            set;
        }

        #endregion Properties

        #region Internal Methods

        /// <summary>
        /// Sends the actual message to individual client
        /// </summary>
        /// <param name="clientID">Client ID where message is sent to</param>
        /// <param name="client">client message to be sent to</param>
        /// <param name="message">message to be sent</param>
        internal void sendMessage(string clientID, TcpClient client, Message message)
        {
            if (client.Connected)
            {
                //update last sent time
                lock (_clientsDictionary)
                {
                    if (_clientsDictionary.ContainsKey(clientID))
                        _clientsDictionary[clientID].LastSent = DateTime.Now;
                }

                if (String.IsNullOrEmpty(message.ClientID))
                    message.SetClientID("SERVER");

                BinaryFormatter bFormatter = new BinaryFormatter();
                bFormatter.Serialize(client.GetStream(), message);
            }
        }

        /// <summary>
        /// Sends a message to an individual client based on logged on username
        /// </summary>
        /// <param name="message">Message to be sent to use</param>
        /// <param name="user">Name of user message is to be sent to</param>
        /// <param name="ignoreCase">Inidcates wether the case of the username should be </param>
        internal void sendMessage(Message message, string user, bool ignoreCase = true)
        {
            int sendCount = 0;

            //copy the clients dictionary in case its modified whilst iterating
            Dictionary<string, ConnectedClient> clients = new Dictionary<string, ConnectedClient>(_clientsDictionary);

            foreach (KeyValuePair<string, ConnectedClient> entry in clients)
            {
                ConnectedClient connClient = null;

                connClient = (ConnectedClient)entry.Value;

                if ((ignoreCase && (connClient.LoginName.ToLower() == user.ToLower())) || (connClient.LoginName == user))
                {
                    sendMessage(connClient.ClientID, _clientsDictionary[connClient.ClientID].Client, message);
                    sendCount++;
                }
            }

            if (sendCount == 0)
                sendMessage(message.ClientID, _clientsDictionary[message.ClientID].Client, new Message("USER_NOT_FOUND", user, MessageType.Error));
        }

        #endregion Internal Methods

        #region Private Methods

        /// <summary>
        /// Retrieves the clients ip address
        /// </summary>
        /// <param name="client">client who's ip address sought</param>
        /// <returns>ip address of client</returns>
        private string ClientAddress(TcpClient client)
        {
            return (((IPEndPoint)client.Client.RemoteEndPoint).Address.ToString());
        }

        /// <summary>
        /// Processes internal command messages from clients
        /// </summary>
        /// <param name="message">Message to process</param>
        /// <returns>true if execution can continue, otherwise false (we handled it)</returns>
        private bool ProcessCommand(Message message)
        {
            bool Result = true;
            TcpClient client = _clientsDictionary[message.ClientID].Client;


            if (String.IsNullOrEmpty(message.ClientID))
                return (false);

            switch (message.Title)
            {
                case "CLOSING":  //Client is about to disconnect
                    Result = false;

                    //if key not found then stop
                    if (!_clientsDictionary.ContainsKey(message.ClientID))
                        return (Result);

                    RaiseClientDisconnected(ClientAddress(client), message.ClientID);

                    break;
                case "LOGIN":  // Client is about to login
                    Result = false;
                    string[] loginDetails = message.Contents.Split('$');

                    if ((loginDetails[0].Length == 0 || loginDetails[1].Length == 0) ||
                        !RaiseOnLogin(ClientAddress(client), loginDetails[0], StringCipher.Decrypt(loginDetails[1], _clientsDictionary[message.ClientID].RandomPassword)))
                    {
                        sendMessage(message.ClientID, client, Message.Command("LOGIN_FAILED"));
                        RaiseClientDisconnected(ClientAddress(client), message.ClientID);
                    }
                    else
                    {
                        _clientsDictionary[message.ClientID].LoggedIn = true;
                        _clientsDictionary[message.ClientID].LoginName = loginDetails[0];
                        sendMessage(message.ClientID, client, Message.Command("LOGGED_IN"));
                    }

                    break;
                case "IGNORE_BROADCAST":
                    _clientsDictionary[message.ClientID].IgnoreBroadcastMessages = Convert.ToBoolean(message.Contents);
                    break;
                case "SET_BUFFER_SIZE":
                    _clientsDictionary[message.ClientID].BufferSize = (FileBufferSize)Convert.ToUInt32(message.Contents);
                    break;
                default:
                    RaiseClientCommand(message);
                    break;
            }


            return (Result);
        }

        /// <summary>
        /// Thread body for listening for client connections
        /// </summary>
        private void ListenForClientConnections()
        {
            while (_running)
            {
                try
                {
                    TcpClient connectedTcpClient = _tcpListener.AcceptTcpClient();

                    lock (this)
                    {
                        // assign next client id
                        activeNewClientID = String.Format("Client:{0}", _nextClientID);

                        while (_clientsDictionary.ContainsKey(activeNewClientID))
                        {
                            _nextClientID += 1;
                            activeNewClientID = String.Format("Client:{0}", _nextClientID);
                        }

                        try
                        {
                            // are there too many clients connected.
                            if (_maxClientConnections > 0 && (ConnectedClientCount() >= _maxClientConnections))
                            {
                                sendMessage(activeNewClientID, connectedTcpClient, new Message("Error", "Too Many Clients", MessageType.Error));
                                connectedTcpClient.Close();
                                continue;
                            }

                            //is the client allowed to connect
                            if (RaiseAllowClientConnect(ClientAddress(connectedTcpClient)))
                            {
                                ConnectedClient client = new ConnectedClient(activeNewClientID, connectedTcpClient);

                                lock (_clientsDictionary)
                                {
                                    _clientsDictionary.Add(activeNewClientID, client);
                                }

                                // Remember the new connection
                                Thread newClientListenThread = new Thread(new ThreadStart(ListenForMessage));
                                client.ListenThread = newClientListenThread;
                                newClientListenThread.Start();
                                newClientListenThread.Name = String.Format("Client Connection - {0}", activeNewClientID);

                                sendMessage(activeNewClientID, connectedTcpClient, Message.Command("NEWCLIENTID", activeNewClientID));

                                RaiseClientConnected(ClientAddress(connectedTcpClient), activeNewClientID);

                                if (_loginRequired)
                                    sendMessage(activeNewClientID, connectedTcpClient, Message.Command("LOGIN_REQUIRED", client.RandomPassword));
                            }
                            else
                            {
                                //not allowed to connect
                                sendMessage(activeNewClientID, connectedTcpClient, Message.Command("CONNECTION_REFUSED"));
                                connectedTcpClient.Close();
                            }

                            //increment next connection id
                            ++_nextClientID;
                        }
                        catch (Exception err)
                        {
                            if (!HandleClientException(err, activeNewClientID, connectedTcpClient))
                                throw;
                        }
                    }
                }
                catch (Exception error)
                {
                    if (!error.Message.Contains("A blocking operation was interrupted by a call to WSACancelBlockingCall"))
                        if (_running)
                            if (!HandleClientException(error, "SERVER", null))
                                throw;
                }
            }
        }

        /// <summary>
        /// Method that performs maintenance for the server
        /// </summary>
        private void MaintenanceThread()
        {
            while (_running)
            {
                if (_clientTimeout > 0)
                {
                    //copy the clients dictionary in case its modified whilst iterating
                    Dictionary<string, ConnectedClient> clients = new Dictionary<string, ConnectedClient>(_clientsDictionary);

                    foreach (KeyValuePair<string, ConnectedClient> entry in clients)
                    {
                        if (!entry.Value.Client.Connected)
                            continue;

                        TimeSpan spanSent = DateTime.Now - entry.Value.LastReceived;
                        TimeSpan spanReceived = DateTime.Now - entry.Value.LastSent;

                        if (spanSent.Seconds > _clientTimeout && spanReceived.Seconds > _clientTimeout)
                        {
                            lock (_clientsDictionary)
                            {
                                sendMessage(entry.Key, entry.Value.Client, Message.Command("TIMEOUT"));

                                if (_clientsDictionary.ContainsKey(entry.Key))
                                    _clientsDictionary[entry.Key].Client.Close();
                            }
                        }
                    }
                }

                Thread.Sleep(1000);
            }
        }

        /// <summary>
        /// Thread method for listening for messages from the clients
        /// </summary>
        private void ListenForMessage()
        {
            ConnectedClient connectedClient = _clientsDictionary[activeNewClientID];
            try
            {
                while (_running)
                {
                    // Block until an instance Message is received
                    BinaryFormatter bFormatter = new BinaryFormatter();
                    Message message = (Message)bFormatter.Deserialize(connectedClient.Client.GetStream());

                    connectedClient.LastReceived = DateTime.Now;

                    //if client is no longer connected then exit
                    if (!connectedClient.Client.Connected)
                        return;

                    // Notify all registered event handlers about the message.
                    if (message != null)
                    {
                        switch (message.Type)
                        {
                            case MessageType.Command:
                                //Command message, are we intercepting and processing it?  if not send it on.
                                if (!ProcessCommand(message))
                                    continue;

                                break;
                            case MessageType.File:
                                ProcessFileRequests(message);

                                continue;
                            case MessageType.User:
                                //User messages handled internally, process and continue
                                sendMessage(message, message.Title);

                                continue;
                            default:
                                //all other messages sent direct to host application

                                //if its a broadcast message then re-broadcast to all clients
                                if (message.Type == MessageType.Broadcast)
                                    SendMessage(message);

                                break;
                        }
                    }

                    if (MessageReceived != null && message != null)
                    {
                        MessageReceived(this, message);
                    }
                }
            }
            catch (Exception err)
            {
                //if the error is not handled then pass it on
                if (!HandleClientException(err, connectedClient.ClientID, connectedClient.Client))
                    throw;
            }
        }

        private void ProcessFileRequests(Message message)
        {
            string newFileName = String.Empty;

            switch (message.Title)
            {
                case "REQUEST_FILE":
                    newFileName = RaiseFileReceive(message.Contents, message.ClientID);
                    FileInfo fi = new FileInfo(newFileName);

                    Message fileMessage = Message.FileToStringMessage(newFileName);
                    sendMessage(message.ClientID, _clientsDictionary[message.ClientID].Client, fileMessage);
                    break;
                default:
                    //File messages are handled internally, process and continue
                    try
                    {
                        newFileName = RaiseFileReceive(message.Title, message.ClientID);

                        byte[] rebin = Convert.FromBase64String(message.Contents);
                        using (FileStream fs = new FileStream(newFileName, FileMode.Create))
                        {
                            using (BinaryWriter bw = new BinaryWriter(fs))
                                bw.Write(rebin);
                        }

                        RaiseFileReceived(this, new TransferFileEventArgs(newFileName, (ulong)message.Contents.Length, (ulong)message.Contents.Length, new TimeSpan(), 0.0, message.ClientID));
                        MessageReceived(this, message);
                        message.Contents = String.Empty;
                        message.Type = MessageType.Acknowledge;
                        sendMessage(message.ClientID, _clientsDictionary[message.ClientID].Client, message);
                    }
                    catch (Exception err)
                    {
                        HandleClientException(err, message.ClientID, _clientsDictionary[message.ClientID].Client);
                        sendMessage(message.ClientID, _clientsDictionary[message.ClientID].Client, new Message(err.Message, err.StackTrace.ToString(), MessageType.Error));
                    }

                    break;
            }
        }

        /// <summary>
        /// Handles exceptions when handling client connections
        /// </summary>
        /// <param name="error">Exception raised</param>
        /// <param name="clientID">clientID when error raised</param>
        /// <returns>bool, true if exception handled, otherwise false</returns>
        private bool HandleClientException(Exception error, string clientID, TcpClient client)
        {
            bool Result = true;

            if (!client.Connected)
            {
                RaiseClientDisconnected(_clientsDictionary[clientID].ClientIP, clientID);
                return (Result);
            }

            if (error.Message == "End of Stream encountered before parsing was completed.")
            {
                //client already disconnected
                RaiseClientDisconnected(ClientAddress(client), clientID);
            }
            else
                if (error.Message.Contains("An existing connection was forcibly closed by the remote host"))
                {
                    //remove the client from the list as its disconnected
                    Message msg = Message.Command("CLOSING");
                    msg.SetClientID(clientID);
                    ProcessCommand(msg);
                    RaiseClientDisconnected(ClientAddress(client), clientID);
                }
                else
                {
                    if (error.Message.Contains("An item with the same key has already been added."))
                    {
                        _nextClientID += 10;
                    }
                    else
                    {
                        //something else happened?
                        if (!RaiseError(error))
                        {
                            RaiseStopped();
                            Result = false;
                        }
                    }
                }

            return (Result);
        }

        /// <summary>
        /// Determines how many clients are currently connected
        /// </summary>
        /// <returns></returns>
        private ushort ConnectedClientCount()
        {
            ushort Result = 0;

            lock (_clientsDictionary)
            {
                foreach (KeyValuePair<string, ConnectedClient> kvp in _clientsDictionary)
                {
                    if (kvp.Value.Client.Connected)
                        Result++;
                }
            }

            return (Result);
        }

        #endregion Private Methods

        #region Events

        #region Private Methods

        /// <summary>
        /// Raises Internal error event handler
        /// </summary>
        /// <param name="error">Error that occured</param>
        /// <returns>bool true to continue false to stop the server</returns>
        private bool RaiseError(Exception error)
        {
            ErrorEventArgs args = new ErrorEventArgs(error);

            if (OnError != null)
                OnError(this, args);

            return (args.Continue);
        }

        /// <summary>
        /// Raises event to indicates a client want to connects
        /// </summary>
        /// <param name="ipAddress"></param>
        /// <returns></returns>
        private bool RaiseAllowClientConnect(string ipAddress)
        {
            ClientAllowConnectEventArgs args = new ClientAllowConnectEventArgs(ipAddress);

            if (AllowClientConnect != null)
                AllowClientConnect(this, args);

            return (args.Allow);
        }

        /// <summary>
        /// Raises the ClientConnected event handler
        /// </summary>
        /// <param name="ClientID">ClientID of client who is connecting</param>
        private void RaiseClientConnected(string ipAddress, string clientID)
        {
            if (ClientConnected != null)
                ClientConnected(this, new ClientEventArgs(ipAddress, clientID));
        }

        /// <summary>
        /// Raised the ClientDisconnected event handler
        /// </summary>
        /// <param name="ClientID">ClientID of client who is disconnecting</param>
        private void RaiseClientDisconnected(string IPAddress, string ClientID)
        {
            if (ClientDisconnected != null)
                ClientDisconnected(this, new ClientEventArgs(IPAddress, ClientID));
        }

        /// <summary>
        /// Raises the Started Event
        /// </summary>
        private void RaiseStarted()
        {
            _running = true;

            if (Started != null)
                Started(this, EventArgs.Empty);
        }

        /// <summary>
        /// Raises the stopped event
        /// </summary>
        private void RaiseStopped()
        {
            _running = false;

            if (Stopped != null)
                Stopped(this, EventArgs.Empty);
        }

        /// <summary>
        /// Raises Client Command event
        /// </summary>
        /// <param name="message"></param>
        private void RaiseClientCommand(Message message)
        {
            if (ClientCommand != null)
                ClientCommand(this, message);
        }

        /// <summary>
        /// Raises an event to indicate the client wants to login
        /// </summary>
        /// <param name="IPAddress">IP Address of client</param>
        /// <param name="UserName">Username provided by client</param>
        /// <param name="Password">Password provided by client</param>
        /// <returns></returns>
        private bool RaiseOnLogin(string IPAddress, string UserName, string Password)
        {
            ClientLoginArgs args = new ClientLoginArgs(IPAddress, UserName, Password);

            if (ClientLogin != null)
                ClientLogin(this, args);

            return (args.LoggedIn);
        }

        /// <summary>
        /// Raise a file received event
        /// 
        /// Contains progess of file received/sent
        /// </summary>
        /// <param name="sender">sender object</param>
        /// <param name="e">ReceiveFileEventArgs</param>
        private void RaiseFileReceived(object sender, TransferFileEventArgs e)
        {
            if (FileReceived != null)
                FileReceived(this, e);
        }

        #endregion Private Methods

        #region Internal Methods

        /// <summary>
        /// Raises a FileReceive event
        /// </summary>
        /// <param name="FileName">Filename being received</param>
        /// <param name="ClientID">ID of client requesting file transfer</param>
        /// <returns>Client should return a fully qualified path/file name of the file to be returned</returns>
        internal string RaiseFileReceive(string FileName, string ClientID)
        {
            TransferFileEventArgs args = new TransferFileEventArgs(FileName, ClientID);

            if (FileReceive != null)
                FileReceive(this, args);

            return (args.FileName);
        }

        #endregion Internal Methods

        public event MessageReceivedEventHandler ClientCommand;
        public event MessageReceivedEventHandler MessageReceived;
        public event ClientAllowConnectEventHandler AllowClientConnect;
        public event ClientEventHandler ClientConnected;
        public event ClientEventHandler ClientDisconnected;
        public event EventHandler Started;
        public event EventHandler Stopped;
        public event ErrorEventHandler OnError;
        public event ClientLoginHandler ClientLogin;
        public event FileReceivedHandler FileReceive;
        public event FileReceivedHandler FileReceived;

        #endregion Events
    }
}
